# -*- coding: utf-8 -*-

# Authors: Natalia B Bidart <natalia.bidart@canonical.com>
#
# Copyright 2010 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The test suite for the GTK UI for the control panel for Ubuntu One."""

from collections import defaultdict

from ubuntuone.controlpanel.gtk import gui
from ubuntuone.controlpanel.gtk.tests.test_package_manager import (
    FakedTransaction)


FAKE_ACCOUNT_INFO = {'type': 'Payed', 'name': 'Test me',
    'email': 'test.com', 'quota_total': '12345', 'quota_used': '9999'}

USER_HOME = '/home/tester'

ROOT = {
    u'volume_id': '', u'path': '/home/tester/My Ubuntu',
    u'subscribed': 'True', u'type': u'ROOT',
}

FAKE_FOLDERS_INFO = [
    {u'volume_id': u'0', u'path': u'/home/tester/foo',
     u'suggested_path': u'~/foo', u'subscribed': u'', u'type': u'UDF'},
    {u'volume_id': u'1', u'path': u'/home/tester/bar',
     u'suggested_path': u'~/bar', u'subscribed': u'True', u'type': u'UDF'},
    {u'volume_id': u'2', u'path': u'/home/tester/baz',
     u'suggested_path': u'~/baz', u'subscribed': u'True', u'type': u'UDF'},
]

FAKE_SHARES_INFO = [
    {u'volume_id': u'1234', u'name': u'do',
     u'path': u'/home/tester/.local/share/ubuntuone/shares/do from Other User',
     u'subscribed': u'', u'type': u'SHARE'},

    {u'volume_id': u'5678', u'name': u're',
     u'path': u'/home/tester/.local/share/ubuntuone/shares/re from Other User',
     u'subscribed': u'True', u'type': u'SHARE'},
]

FAKE_VOLUMES_INFO = [
    (u'', u'147852369', [ROOT] + FAKE_FOLDERS_INFO),
    (u'Other User', u'985674', FAKE_SHARES_INFO),
]

FAKE_DEVICE_INFO = {
    'device_id': '1258-6854', 'device_name': 'Baz', 'device_type': 'Computer',
    'is_local': 'True', 'configurable': 'True', 'limit_bandwidth': 'True',
    'max_upload_speed': '1000', 'max_download_speed': '72548',
}

FAKE_DEVICES_INFO = [
    {'device_id': '0', 'name': 'Ubuntu One @ Foo', 'type': 'Computer',
     'is_local': '', 'configurable': ''},
    {'device_id': '1', 'name': 'Ubuntu One @ Bar', 'type': 'Phone',
     'is_local': '', 'configurable': ''},
    {'device_id': '2', 'name': 'Ubuntu One @ Z', 'type': 'Computer',
     'is_local': '', 'configurable': 'True', 'limit_bandwidth': '',
     'max_upload_speed': '0', 'max_download_speed': '0'},
    {'device_id': '1258-6854', 'name': 'Ubuntu One @ Baz', 'type': 'Computer',
     'is_local': 'True', 'configurable': 'True', 'limit_bandwidth': 'True',
     'max_upload_speed': '1000', 'max_download_speed': '72548'},  # local
]

FAKE_REPLICATIONS_INFO = [
    {'replication_id': 'foo', 'name': 'Bar',
     'enabled': 'True', 'dependency': ''},
    {'replication_id': 'yadda', 'name': 'Foo',
     'enabled': '', 'dependency': 'a very weird one'},
    {'replication_id': 'yoda', 'name': 'Figthers',
     'enabled': 'True', 'dependency': 'other dep'},
]


class FakedObject(object):
    """Fake an object, record every call."""

    exposed_methods = []

    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        self._called = {}
        for i in self.exposed_methods:
            setattr(self, i, self._record_call(i))

    def _record_call(self, func_name):
        """Store values when calling 'func_name'."""

        def inner(*args, **kwargs):
            """Fake 'func_name'."""
            self._called[func_name] = (args, kwargs)

        return inner


class FakedNMState(FakedObject):
    """Fake a NetworkManagerState."""

    exposed_methods = ['find_online_state']


class FakedDBusBackend(FakedObject):
    """Fake a DBus Backend."""

    bus_name = None
    object_path = None
    iface = None

    def __init__(self, obj, dbus_interface, *args, **kwargs):
        if dbus_interface != self.iface:
            raise TypeError()
        self._signals = defaultdict(list)
        super(FakedDBusBackend, self).__init__(*args, **kwargs)

    def connect_to_signal(self, signal, handler):
        """Bind 'handler' to be callback'd when 'signal' is fired."""
        self._signals[signal].append(handler)


class FakedSSOBackend(FakedDBusBackend):
    """Fake a SSO Backend, act as a dbus.Interface."""

    bus_name = gui.ubuntu_sso.DBUS_BUS_NAME
    object_path = gui.ubuntu_sso.DBUS_CREDENTIALS_PATH
    iface = gui.ubuntu_sso.DBUS_CREDENTIALS_IFACE
    exposed_methods = ['find_credentials', 'clear_credentials',
                       'login', 'register']


class FakedControlPanelBackend(FakedDBusBackend):
    """Fake a Control Panel Backend, act as a dbus.Interface."""

    bus_name = gui.DBUS_BUS_NAME
    object_path = gui.DBUS_PREFERENCES_PATH
    iface = gui.DBUS_PREFERENCES_IFACE
    exposed_methods = [
        'account_info',  # account
        'devices_info', 'change_device_settings', 'remove_device',  # devices
        'volumes_info', 'change_volume_settings',  # volumes
        'replications_info', 'change_replication_settings',  # replications
        'file_sync_status', 'enable_files', 'disable_files',  # files
        'connect_files', 'disconnect_files',
        'restart_files', 'start_files', 'stop_files',
    ]


class FakedSessionBus(object):
    """Fake a session bus."""

    def get_object(self, bus_name, object_path, introspect=True,
                   follow_name_owner_changes=False, **kwargs):
        """Return a faked proxy for the given remote object."""
        return None


class FakedInterface(object):
    """Fake a dbus interface."""

    def __new__(cls, obj, dbus_interface, *args, **kwargs):
        if dbus_interface == gui.DBUS_PREFERENCES_IFACE:
            return FakedControlPanelBackend(obj, dbus_interface,
                                            *args, **kwargs)
        if dbus_interface == gui.ubuntu_sso.DBUS_CREDENTIALS_IFACE:
            return FakedSSOBackend(obj, dbus_interface, *args, **kwargs)


class FakedPackageManager(object):
    """Faked a package manager."""

    def __init__(self):
        self._installed = {}
        self.is_installed = lambda package_name: \
            self._installed.setdefault(package_name, False)

    @gui.package_manager.inline_callbacks
    def install(self, package_name):
        """Install 'package_name' if is not installed in this system."""
        yield
        self._installed[package_name] = True
        gui.package_manager.return_value(FakedTransaction([package_name]))


class FakedConfirmDialog(object):
    """Fake a confirmation dialog."""

    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        self.was_run = False
        self.is_visible = True
        self.show = lambda: setattr(self, 'is_visible', True)
        self.hide = lambda: setattr(self, 'is_visible', False)
        self.response_code = None

    def run(self):
        """Set flag and return 'self.response_code'."""
        self.was_run = True
        return self.response_code
